unit Music;

{
   MUSIC.PAS vam umoznuje hrat hudbu na vasem IBM PC nebo kompatibilnim
   pocitaci s pouzitim stejne sady prikazu, jako pouziva prikaz "PLAY"
   v BASICu.

   Puvodni modul byl napsan Gragory Arakelianem (74017,223) 703-435-7137.
   Do jednotky Turbo Pascalu ho prepsal Ted Lassagne (70325,206).
   Kod byl rozsiren o manipulaci s notami. Byly pridelany nektere
   osetreni chyb spolu s malymi upravami a optimalizaci kodu.
}

{=======================================================================}

interface

uses CRT;

Procedure Play (TuneString:string);

  {Play interpretuje retezec velmi podobne jako PLAY v BASICu. Jsou
   pouze dve vetsi odlisnosti. "N" neni interpretovano a promenne
   se nemohou vyskytovat v retezci.

   Znaky v retezci jsou interpretovany takto:

      A .. G    Hudebni noty A..G. Nota A muze byt nasledovana tzv.
                "posuvkou" (# nebo '+' tzn. o pulton vyssi resp. nizsi)
                Nota muze byt nasledovana cislem oznacujicim delku
                noty. (1 pro celou notu do 64 pro sedesatictyrctvrtinovou
                notu) Nota s volitelnou posuvkou a delkou meze byt nasledovana
                tez jednou nebo vice teckami. Kazda tecka prodluzuje notu
                o polovinu jeji dosavadni delky. Napriklad, dve tecky
                prodluzuji notu na 9/4 puvodni delky a tri tecky
                na 27/8 puvodni delky.

      Ln        Specifikuje preddef. delku noty nasledujicim zpusobem:
                ("n" musi byt 1 pro celou notu az 64 pro 64-tinovou
                notu) Inicializacni hodnota je 4 (ctvrtinova nota).

      Mz        Oznacuje zlomek delky noty, ve kterem nota zni.
                "z" je jedno z pismen "S", "N", nebo "L", ktera
                maji tento vyznam:

                   MS   staccato   (3/4 delka)
                   MN   normalni   (7/8 delka)
                   ML   legato     (cela delka)

      On        Oznacuje oktavu, ve ktere jsou nasledujici noty
                hrany (od 0 do 7). Preddef. hodnota je 3.

      Pn        Specifikuje dobu, po jakou se nebude hrat zadny ton.
                "n" je delka noty (1 pro celou delku noty az do 64-tinove
                noty). Jestlize je "n" vynechano, pouzije se aktalni
                preddef. delka. Muze nasledovat jedna nebo vice tecek.
                Kazda z nich prodlouzi "dobu ticha" o jednu polovinu
                jeji dosavadni hodnoty.

      Tn        Oznacuje tempo v "uderech" za minutu (od 32 do
                255.)  Preddef. hodnota je 120.

      Poznamka: Hra muze byt prerusena v jakemkoli okamziku stlacenim
      Control-C nebo Control-Break. To ukonci program a vrati se do DOSu.
      Jestlize toto chcete zmenit, ridici kod pro klavesnici nasleduje
      hned za kodem noty.

}

{=======================================================================}

implementation


Const
    SharpOffset = 60;

Var
    PitchArray : Array[1..120] Of Integer;
      {Prvnich 56 zaznamu v PitchArray jsou frekvence pro
       noty A..G v sedmi oktavach. Zaznamy 60 az 115 jsou
       frekvence pro noty o pul tonu vyssi v prvnich 56 zaznamech}
    BaseOctave : Integer;
    Octave     : Integer;
    GenNoteType: Integer;
    Tempo      : Integer;
    PlayFrac   : Byte;


{PlayInit nastavuje preddef. hodnoty pro oktavy, delku noty, tempo
a modifikatory delky. Inicializuje pole frekvenci pro noty}

Procedure PlayInit;
  Const
      NextFreq    = 1.05946309436;
  Var
      RealFreq : Array[1..7] Of Real;
      BaseFreq : Real;
      J,K      : Integer;
  Begin

   {Nastaveni preddef. hodnot}

    BaseOctave := 0;
    Octave := 3;         {Treti oktava}
    GenNoteType := 4;    {Ctvrtinova nota}
    Tempo := 120;        {120 "uderu" za minutu}
    PlayFrac := 7;       {Normalni - noty zni 7/8 celk.casu}

    {Nastaveni pole frekvenci}

    BaseFreq := 27.5;    {"A" ctyri oktavy pod A-440}
    For J := 0 To 7 Do
      Begin
        RealFreq[1] := BaseFreq;
        RealFreq[2] := RealFreq[1]*NextFreq*NextFreq;
        RealFreq[3] := RealFreq[2]*NextFreq;
        RealFreq[4] := RealFreq[3]*NextFreq*NextFreq;
        RealFreq[5] := RealFreq[4]*NextFreq*NextFreq;
        RealFreq[6] := RealFreq[5]*NextFreq;
        RealFreq[7] := RealFreq[6]*NextFreq*NextFreq;
        BaseFreq := BaseFreq * 2;   {dalsi oktava}
        For K := 1 to 7 Do
          Begin
            PitchArray[J*7+K] := Round(RealFreq[K]);
            PitchArray[J*7+K+SharpOffset] := Round(RealFreq[K]*NextFreq);
          End;
      End;
  End;


{Play interpretuje zadany retezec a hraje specifikovane noty po zadane
 doby}

Procedure Play (TuneString:string);
  Var PlayTime,IdleTime,DotTime,NoteTime  : Integer;
      NoteType,PitchIndex,Position,Number : Integer;
      Code,TuneStrLen                     : Integer;
      Character                           : Char;

  Procedure NVal(Pos:integer; var v, code: integer);
  {Vybere hodnotu "v" z retezce pocinajiciho na Pos. Vracena hodnota
  "code" je pocet sejmutych cislic plus jedna.}
     var  posn:integer;
     begin
        v := 0;
        posn := Pos;
        while (posn <= TuneStrLen) and
        (TuneString[posn] in ['0'..'9']) do begin
           v := v*10 + ord(TuneString[posn]) - ord ('0');
           posn := posn + 1;
        end;
        code := posn - Pos + 1;
     end {NVal};

  Procedure CheckDots;
  {Kontroluje pocet tecek po note nebo pauze. Kazda tecka zvysuje
   delku noty nebo ticha o polovinu}
    begin
       while (Position <= TuneStrLen) and
       (TuneString[Position] = '.') do begin
          DotTime := DotTime + DotTime div 2;
          inc(Position)
       end;
    end {CheckDots};

  Begin {podprogram Play}
    CheckBreak := false;
    TuneStrLen := length(TuneString);
    Position := 1;

    Repeat
      NoteType := GenNoteType;
      DotTime := 1000;

      Character := upcase(TuneString[Position]);
      Case Character Of
        'A'..'G' : Begin
                     PitchIndex := (ord(Character)-64)+Octave*7;
                     If (Character='A') or (Character='B') Then
                       PitchIndex := PitchIndex + 7;  {dalsi oktava}
                     inc(Position);

                     {Test pro o pul tonu vyssi resp. nizsi notu}
                     if Position <= TuneStrLen then
                        case TuneString[Position] of
                          '#','+': begin
                            PitchIndex := PitchIndex+SharpOffset;
                            inc(Position);
                           end;
                          '-': begin
                            PitchIndex := PitchIndex+SharpOffset - 1;
                            inc(Position);
                           end;
                        End;

                     {Testuje delku nasledujici noty}
                     if (Position <= TuneStrLen) and
                     (TuneString[Position] in ['0'..'9']) then begin
                        NVal(Position,NoteType,Code);
                        inc(Position, Code - 1)
                     end;

                     {Testuje tecky, ktere nasleduji notu}
                     CheckDots;

                     {Hraje notu}
                     NoteTime := Round(DotTime/Tempo/NoteType*240);
                     PlayTime := Round(NoteTime*PlayFrac/8);
                     IdleTime := NoteTime-PlayTime;
                     Sound(PitchArray[PitchIndex]);
                     Delay(PlayTime);
                     if IdleTime <> 0 then begin
                        NoSound;
                        Delay(IdleTime)
                     end;

                     {Test, zda-li je Ctrl-Break}
                     if keypressed and (ReadKey = ^C) then begin
                        NoSound;
                        halt
                     end;

                   End;
             'L' :  {Delka noty (1 az 64).  "1" znamena
                     celou notu a "64" 64-tictvrtinovou notu.}
                   Begin
                     NVal (Position+1,GenNoteType,Code);
                     if (GenNoteType < 1) or (GenNoteType > 64) then
                        GenNoteType := 4;
                     inc(Position, Code);
                   End;
             'M' :  {Modifikace delky noty - "S" znamena staccato,
                     "L" znamena legato, nebo "N" znamena normalni.}
                   Begin
                     if Position < TuneStrLen then begin
                        Case upcase(TuneString[Position+1]) Of
                          'S' : PlayFrac := 6;
                          'N' : PlayFrac := 7;
                          'L' : PlayFrac := 8;
                        End;
                        inc(Position, 2);
                     end;
                   End;
             'O' :  {Specifikace oktavy (0 az 7)}
                   Begin
                     NVal (Position+1,Octave,Code);
                     Octave := Octave+BaseOctave;
                     if Octave > 7 then Octave := 3;
                     inc(Position, Code);
                   End;
             'P' :  {Pauza (rest), ktera muze nabyvat hodnoty 1 az 64}
                    Begin
                     NoSound;
                     NVal (Position+1,NoteType,Code);
                     if (NoteType < 1) or (NoteType > 64) then
                        NoteType := GenNoteType;
                     inc(Position, Code);
                     CheckDots;
                     IdleTime := DotTime Div Tempo * (240 Div NoteType);
                     Delay (IdleTime);
                   End;
             'T' :  {Tempo - pocet "uderu" za minutu (32 - 255)}
                   Begin
                     NVal (Position+1,Tempo,Code);
                     if (Tempo < 32) or (Tempo > 255) then
                        Tempo := 120;
                     inc(Position, Code);
                   End;
            Else inc(Position);   {Ignoruje neplatne znaky}
      End;
    Until Position > TuneStrLen;
    NoSound;
  End {Play};

Begin    {Inicializace}

  PlayInit;

End.