{Polarwolf Library V2.2}

Procedure EnterGraphMode;
  Begin
    If InGraphMode Then Exit;
    If Not CopyRightRequested Then RunError;
    InitGraph (GraphDriver, GraphMode, '');
    If GraphDriver >= 0 Then
      Begin
        InGraphMode := True;
        SetColor (StdColor);
        SetBkColor (StdBackground)
      End
  End;

Procedure LeaveGraphMode;
  Begin
    If Not InGraphMode Then Exit;
    CloseGraph;
    InGraphMode := False
  End;

Function MaxXPos : Byte;
  Begin
    MaxXPos := Lo (WindMax) - Lo (WindMin) +1
  End;

Function MaxYPos : Byte;
  Begin
    MaxYPos := Hi (WindMax) - Hi (WindMin) +1
  End;

Function GrCorX (Cor : Byte) : Word;
  Begin
    Cor := Cor + Lo (WindMin) -1;
    If Cor > Lo (WindMax) Then Cor := Lo (WindMax);
    GrCorX  := Cor Shl 3
  End;

Function MaxColums : Byte;
  Begin
    MaxColums := (GetMaxX +1) Shr 3
  End;

Function MaxRows : Byte;
  Begin
    MaxRows := (GetMaxY +1) Shr 3
  End;

Function GrCorY (Cor : Byte) : Word;
  Begin
    Cor := Cor + Hi (WindMin) -1;
    If Cor > Hi (WindMax) Then Cor := Hi (WindMax);
    GrCorY := Cor Shl 3
  End;

Procedure SaveColor;
  Begin
    LastColorInfo := GetColor;
    GetFillSettings (LastFillInfo);
    GetTextSettings (LastTextInfo);
    GetViewSettings (LastViewPortInfo);
    If GetMaxColor < 15 Then
      Begin
        If TextAttr And (16+32+64) > 0 Then
          Begin
            SetFillStyle (SolidFill, 1);
            SetColor (0)
          End
          Else Begin
            SetFillStyle (SolidFill, 0);
            If (TextAttr And (15) > 0) Then SetColor (1)
              Else SetColor (0)
          End;
      End
      Else Begin
        SetFillStyle (SolidFill, TextAttr Shr 4);
        SetColor (TextAttr And 15)
      End;
    SetTextStyle (DefaultFont, HorizDir, 1);
    SetTextJustify (LeftText, TopText);
    SetViewPort (0, 0, GetMaxX, GetMaxY, ClipOn)
  End;

Procedure RestoreColor;
  Begin
    SetColor (LastColorInfo);
    SetFillStyle (LastFillInfo.Pattern, LastFillInfo.Color);
    SetTextStyle (LastTextInfo.Font, LastTextInfo.Direction, LastTextInfo.CharSize);
    SetTextJustify (LastTextInfo.Horiz, LastTextInfo.Vert);
    SetViewPort (LastViewPortInfo.X1, LastViewPortInfo.Y1,
      LastViewPortInfo.X2, LastViewPortInfo.Y2, LastViewPortInfo.Clip)
  End;

Procedure SetChar (Ch : Char);
  Begin
    Bar (GrCorX (XCor), GrCorY (YCor), GrCorX (XCor) + 7, GrCorY (YCor) +7);
    OutTextXY (GrCorX (XCor), GrCorY (YCor), Ch)
  End;

Procedure ScrollLine (ScrollSource, ScrollDest : Byte; Buffer : Pointer);
  Begin
    If ScrollSource > Hi (WindMax) - Hi (WindMin) + 1 Then Exit;
    If ScrollDest > Hi (WindMax) - Hi (WindMin) + 1 Then Exit;
    GetImage (Lo (WindMin) Shl 3,
              (Hi (WindMin) + ScrollSource -1) Shl 3,
              Lo (WindMax) Shl 3 +7,
              (Hi (WindMin) + ScrollSource -1) Shl 3 +7, Buffer^);
    PutImage (Lo (WindMin) Shl 3,
              (Hi (WindMin) + ScrollDest -1) Shl 3, Buffer^, NormalPut)
  End;

Procedure Scroll (ScrollStart, ScrollEnd, ScrollDest : Byte);
  Var Buffer : Pointer;
      BufferSize : Word;
      I : Byte;
  Begin
    BufferSize := ImageSize (0, 0, GetMaxX, 7);
    GetMem (Buffer, BufferSize);
    If ScrollStart > ScrollDest Then
      For I := 0 To ScrollEnd - ScrollStart Do
        ScrollLine (ScrollStart +I, ScrollDest +I, Buffer)
      Else For I := ScrollEnd - ScrollStart -1 DownTo 0 Do
        ScrollLine (ScrollStart +I, ScrollDest +I, Buffer);
    FreeMem (Buffer, BufferSize)
  End;

Procedure ScrollScreen;
  Var Puffer : Pointer;
  Begin
    If GraphDriver <> HERCMono Then Scroll (2, MaxRows, 1)
      Else Begin
        GetMem (Puffer, 32000);
        GetImage (GrCorX (1), GrCorY (2), GrCorX (MaxXPos)+7, GrCorY (MaxRows)+7, Puffer^);
        PutImage (GrCorX (1), GrCorY (1), Puffer^, NormalPut);
        FreeMem (Puffer, 32000);
      End;
    Bar (GrCorX (1), GrCorY (MaxRows), GrCorX (MaxColums) +7, GrCorY (MaxRows) +7)
  End;

Procedure WriteChar (Ch : Char);
  Begin
    Case Ch Of
        #7   : Play (Beep);
        #8   : If XCor > 1 Then Dec (XCor);
        #10  : If YCor < MaxYPos Then Inc (YCor)
                 Else ScrollScreen;
        #13  : XCor := 1
      Else Begin
        SetChar (Ch);
        If XCor < MaxXPos Then Inc (XCor)
          Else If YCor < MaxYPos Then
            Begin
              XCor := 1;
              Inc (YCor)
            End
            Else Begin
              XCor := 1;
              ScrollScreen
            End
        End
      End
  End;

Procedure WriteString (S : String);
  Var I : Byte;
  Begin
    SaveColor;
    For I := 1 To Length (S) Do WriteChar (S [I]);
    RestoreColor
  End;

{$F+}
Function Open43 (Var F : TextRec) : Integer;
  Begin
    If F.Mode = fmOutput Then
      Begin
        If Not InGraphMode Then EnterGraphMode;
        XCor := 1;
        YCor := 1;
        Open43 := 0
      End
    Else Open43 := 12
  End;

Function Out43 (Var F : TextRec) : Integer;
  Var I : Word;
      S : String;
  Begin
    S := '';
    If F.BufPos > 0 Then
      For I := 0 To F.BufPos -1 Do S := S + F.BufPtr^[I];
    F.BufPos := 0;
    If InGraphMode Then WriteString (S);
    Out43 := 0
  End;

Function Close43 (Var F : TextRec) : Integer;
  Begin
    Close43 := 0
  End;
{$F-}

Procedure Assign43 (Var F : Text);
  Begin
    With TextRec (F) Do
      Begin
        Handle  := $FFFF;
        Mode    := fmClosed;
        BufSize := SizeOf (Buffer);
        BufPtr    := @Buffer;
        OpenFunc  := @Open43;
        InOutFunc := @Out43;
        FlushFunc := @Out43;
        CloseFunc := @Close43;
        Name[0] := #0
      End
  End;

Procedure GotoXY (X, Y : Byte);
  Begin
    If InGraphMode Then
      Begin
        If (X <= MaxXPos) And (X > 0) Then XCor := X;
        If (Y <= MaxYPos) And (Y > 0) Then YCor := Y;
      End
      Else CRT.GotoXY (X, Y)
  End;

Procedure ClrEol;
  Begin
    If InGraphMode Then
      Begin
        SaveColor;
        Bar (GrCorX (XCor), GrCorY (YCor),
            GrCorX (MaxXPos) +7, GrCorY (YCor) + 7);
        RestoreColor
      End
      Else CRT.ClrEol
  End;

Procedure ClrScr;
  Begin
    If InGraphMode Then
      Begin
        SaveColor;
        If (WindMin = 0) And (WindMax = MaxColums -1 + (MaxRows -1) Shl 8)
          Then Bar (0, 0, GetMaxX, GetMaxY)
          Else Bar (GrCorX (1), GrCorY (1), GrCorX (MaxXPos) +7, GrCorY (MaxYPos) +7);
        RestoreColor;
        XCor := 1;
        YCor := 1
      End
      Else CRT.ClrScr
  End;

Procedure DelLine;
  Begin
    If InGraphMode Then
      Begin
        SaveColor;
        Scroll (YCor+1, MaxYPos, YCor);
        Bar (GrCorX (1), GrCorY (MaxRows),
             GrCorX (MaxColums) +7, GrCorY (MaxRows) +7);
        RestoreColor
      End
      Else CRT.DelLine
  End;

Procedure InsLine;
  Begin
    If InGraphMode Then
      Begin
        SaveColor;
        Scroll (YCor, MaxYPos, YCor +1);
        Bar (GrCorX (1), GrCorY (YCor),
             GrCorX (MaxColums) +7, GrCorY (YCor) +7);
        RestoreColor
      End
      Else CRT.InsLine
  End;

Function WhereX : Byte;
  Begin
    If InGraphMode Then WhereX := XCor Else WhereX := CRT.WhereX
  End;

Function WhereY : Byte;
  Begin
    If InGraphMode Then WhereY := YCor Else WhereY := CRT.WhereY
  End;

Procedure TextMode (Mode : Integer);
  Begin
    If (Mode And Font8x8 > 0) Then
      Begin
        Assign43 (Output);
        ReWrite (OutPut);
        If GraphDriver < 0 Then Exit;
        SetViewPort (0, 0, GetMaxX, GetMaxY, False);
        Window (1, 1, MaxColums, MaxRows)
      End;
    If (Mode And Font8x8 = 0) And InGraphMode Then
      Begin
        AssignCrt (Output);
        Rewrite (Output);
        LeaveGraphMode;
        Window (1, 1, 80, 25)
      End;
    If Mode And Font8x8 = 0 Then CRT.TextMode (Mode)
  End;

Procedure Window (X1, Y1, X2, Y2 : Byte);
  Var H : Byte;
  Begin
    If InGraphMode Then
      Begin
        If X1 > X2 Then
          Begin
            H := X1;
            X1 := X2;
            X2 := H
          End;
        If Y1 > Y2 Then
          Begin
            H := Y1;
            Y1 := Y2;
            Y2 := H
          End;
        If (X1 >  0) And (X2 >  0) And (Y1 >  0) And (Y2 >  0) And
           (X1 <= MaxColums) And (X2 <= MaxColums) And
           (Y1 <= MaxRows)  And (Y2 <= MaxRows)  Then
           Begin
             WindMin := X1 -1 + (Y1 -1) Shl 8;
             WindMax := X2 -1 + (Y2 -1) Shl 8
           End
      End
    Else CRT.Window (X1, Y1, X2, Y2)
  End;
